# Copyright (C) 2010-2018 by Dr. Dieter Maurer <dieter@handshake.de>
"""schema related utilities."""
from zope.interface import Interface, providedBy, implementer
from zope.interface.interfaces import IMethod
from zope.schema.interfaces import IField, IObject, \
     WrongContainedType, SchemaNotProvided, ValidationError, \
     SchemaNotFullyImplemented
from zope.schema import Object as ObjectBase, Bool, getFieldsInOrder

from .interfaces import ISchemaConfigured

from dm.reuse.proxy import make_proxy


def interface_from_spec(spec, name=None, **kw):
  """build an interface from interface spec *spec*.

  This should probably go into its own package (maybe ``dm.zope.interface``).

  Note: the interfaces constructed in this way may not be picklable
  (not tested). If they are indeed not picklable, they should not be stored.
  """
  # try to guess whether "spec" is an interface or specification
  #  The implementors have broken "issubclass", we therefore need
  #  more indirect (and in principle less reliable) methods
  # if issubclass(spec, Interface):
  if hasattr(spec, 'names'):
    if name is None and not kw: return spec
    spec = (spec,)
  name = name is None and 'FromSpec' or name
  return type(Interface)(name, tuple(spec), kw)


def schemaitems(spec):
  """The schema part of interface specification *spec* as a list of id, field pairs."""
  iface = interface_from_spec(spec)
  # may want to filter duplicates out or raise an exception on duplicates
  seen = set(); items = []
  for (name, field) in getFieldsInOrder(iface):
    if name in seen: continue
    seen.add(name); items.append((name, field))
  return items


def schemadict(spec):
  """The schema part of interface specification *spec* as a ``dict``."""
  return dict(schemaitems(spec))
  

@implementer(ISchemaConfigured)
class SchemaConfigured(object):
  """Mixin class to provide configuration by the provided schema components."""

  def __init__(self, **kw):
    schema = schemadict(self.sc_schema_spec())
    for k in kw:
      # might want to control this check
      if k not in schema:
        raise TypeError('non schema keyword argument: %s' % k)
      setattr(self, k, kw[k])
    # provide default values for schema fields not set
    for f in schema:
      if not hasattr(self, f): setattr(self, f, schema[f].default)

  # provide control over which interfaces define the data schema
  SC_SCHEMAS = None

  def sc_schema_spec(self):
    """the schema specification which determines the data schema.

    This is determined by `SC_SCHEMAS` and defaults to `providedBy(self)`.
    """
    spec = self.SC_SCHEMAS
    if spec is None: return providedBy(self)
    return spec


class SchemaConfiguredEvolution(object):
  """Mixin class to support schema evolution for `SchemaConfigured` subclasses.

  Must come early in the resulting class's `mro` to be effective.
  """
  def __setstate__(self, state):
    sst = getattr(super(SchemaConfigured, self), "__setstate__", None)
    if sst is None and isinstance(state, dict):
      self.__dict__.update(state) # default
    else: sst(state)
    SchemaConfigured.__init__(self) # provide default values for missing fields


class _IObject(IObject):
  """auxiliary interface for ``Object`` extension.

  Note: no support yet for internationalization.
  """
  check_declaration = Bool(
    title=u"check declaration",
    description=u"check that the validated object declares the provided interface",
    default=True,
    )


@implementer(_IObject)
class Object(ObjectBase):
  """improved ``zope.schema.Object``.

  Error reports for ``zope.schema.Object`` forget to specify the
  problematic field. This class adds them.

  Occasionally, we do not wish to check that the existance of an
  interface declaration. Make this controllable with a
  ``check_declaration`` property.
  """

  # work around stupid ``Field`` constructor
  def __init__(self, *args, **kw):
    if 'check_declaration' in kw:
      self.check_declaration = kw['check_declaration']
      del kw['check_declaration']
    super(Object, self).__init__(*args, **kw)

  def _validate(self, value):
    base_validate = super(Object, self)._validate
    validate = base_validate if self.check_declaration \
               else (
      # suppress the declaration check via a proxy
      #   the suppression works be forcing the check to unconditionally succeed
      make_proxy(self,
                 schema=make_proxy(self.schema,
                                   providedBy=lambda unused: True
                                   )
                 )._rebind_proxied_method(base_validate)
      )
    validate(value)
##    super(ObjectBase, self)._validate(value)

##    # schema has to be provided by value
##    if self.check_declaration and not self.schema.providedBy(value):
##      raise SchemaNotProvided(self.__name__)

    # check the value against schema
    # code used in earlier versions -- might still
    #  be necessary for old `zope.schema` versions
##    errors = []
##    for name, field in schemadict(self.schema).items():
##      exc = None
##      try: field.validate(getattr(value, name))
##      except ValidationError as error: exc = error
##      except AttributeError as error: exc = SchemaNotFullyImplemented(error)
##      if exc is not None: errors.append((name, exc))
##    errors = get_validation_errors(self.schema, value)
##    if errors:
##      for n, e in errors:
##        if not hasattr(e, "field"): e.field = n
##        elif n is not None: e.field = n
##      raise WrongContainedType(errors, self.__name__)
  

##from zope.schema._bootstrapfields import get_validation_error



# Copyright (C) 2010-2018 by Dr. Dieter Maurer <dieter@handshake.de>
"""Schema verification.

This is a companion to ``zope.interface.verify`` which verifies methods
but does not do anything with schema fields.
"""

from .schema import Object

def verify_schema(iface, obj, context=None, check_declaration=True):
  """verify that *obj* satisfies the schema requirements of *iface*.

  *context* specifies the context for schema binding. It defaults to *obj*.

  *check_declaration* checks that *obj* declares to provide *iface*.
  """
  if context is None: context = obj
  Object(iface, check_declaration=check_declaration).bind(context).validate(obj)



# Copyright (C) 2011-2018 by Dr. Dieter Maurer <dieter@handshake.de>
from zope.interface import Interface


class ISchemaConfigured(Interface):
  """marker interface for ``SchemaConfigured`` classes.

  Used to facilitate the registration of forms and views.
  """



# Copyright (C) 2011-2018 by Dr. Dieter Maurer <dieter@handshake.de>
"""form support for ``SchemaConfigured`` classes.

This module requires ``zope.formlib``.
"""
from zope.interface import providedBy
from zope.formlib.form import EditForm, DisplayForm, Fields

from .schema import schemaitems


class SchemaConfiguredFormMixin(object):
  """mixin class for ``SchemaConfigured`` forms.

  It redefines ``__init__`` to generate the fields from *context*
  and then calls ``customize_fields`` (if defined) to allow
  for field customization.
  """

  def __init__(self, context, request):
    super(SchemaConfiguredFormMixin, self).__init__(context, request)
    self.define_fields()


  def define_fields(self):
    self.form_fields = Fields(
      *(i[1] for i in schemaitems(self.context.sc_schema_spec()))
      )
    self.customize_fields()


  def customize_fields(self):
    """to be overridden by derived classes."""
    pass


class SchemaConfiguredEditForm(SchemaConfiguredFormMixin, EditForm): pass

class SchemaConfiguredDisplayForm(SchemaConfiguredFormMixin, DisplayForm): pass



# Copyright (C) 2011-2018 by Dr. Dieter Maurer <dieter@handshake.de>
"""Some widget extensions."""
from datetime import timedelta

from zope.i18nmessageid import MessageFactory

from dm.reuse import rebindFunction


_ = MessageFactory('dm_zope_schema')

try: import zope.formlib.textwidgets
except ImportError: modern = False
else: modern = True
  
if modern:
  from zope.formlib.textwidgets import TextWidget, ConversionError, \
       PasswordWidget
  from zope.formlib.widget import DisplayWidget, InputErrors
  from zope.formlib.interfaces import IDisplayWidget, IInputWidget
else:
  from zope.app.form.browser.textwidgets import TextWidget, ConversionError, \
       PasswordWidget
  from zope.app.form.browser.widget import DisplayWidget, InputErrors
  from zope.app.form.interfaces import IDisplayWidget, IInputWidget


##############################################################################
## Timedelta support

def format_timedelta(value):
  if value is None: return ''
  pairs = []
  if value.days: pairs.append("%dd" % value.days)
  if value.seconds: pairs.append("%ds" % values.seconds)
  if value.microseconds: pairs.append("%dm" % values.microsecond)
  if not pairs: pairs.append("0s")
  return " ".join(pairs)


class TimedeltaInputWidget(TextWidget):
  """widget for timedelta input.

  The values are space separated concatenations of sign/int/unit tuples.
  *sign* is missing, `+` or '-'.
  *unit* is `d` (day), `s` (second) and `m` (microsecond).
  Examples are `5d 17s`, `300s`, `-5d 17s`, `-5d -17s`.
  """
  # maps `unit` to target, factor pairs
  #  the factors are there to support e.g. hours, minutes, ...
  CONV = dict(d=("d", 1), s=("s", 1), m=("m", 1))

  def _toFieldValue(self, input):
    value = super(TimedeltaInputWidget, self)._toFieldValue(input.strip())
    if value is None: return value
    if value.startswith("-"): neg = True; value = value[1:]
    else: neg = False
    v = dict(d=0, s=0, m=0); conv = self.CONV
    try:
      for pair in value.split():
        nf = 1
        if pair.startswith("-"): nf = -1; pair = pair[1:]
        elif pair.startswith("+"): pair = pair[1:]
        if not pair:
          raise ValueError(_(u"missing unit", u"missing unit"))
        unit = pair[-1]
        if unit not in conv:
          raise ValueError(_(u"unknown unit", u"unknown unit"))
        t, f = conv[unit]
        if len(pair) == 1: pv = 1 # only unit
        else: pv = int(pair[:-1], 10) # explicitely specified
        v[t] += nf * pv * f
      return timedelta(v["d"], v["s"], v["m"])
    except ValueError, v:
      raise ConversionError(_(u"invalid timedelta", u"invalid timedelta"), v)

  def _toFormValue(self, value): return format_timedelta(value)


class TimedeltaDisplayWidget(DisplayWidget):
  __call__ = rebindFunction(DisplayWidget.__call__,
                            escape=lambda x: "<span>%s</span>" % (x and format_timedelta(x))
                            )


##############################################################################
## Decent `Password` support

class PasswordInputWidget(PasswordWidget):
  """Improved `PasswordWidget`.

  The standard Zope `PasswordWidget` replaces the field value by `''`.
  This forces the user to provide a new value each time the form is
  edited. If he forgets it, the password is lost.

  Our `PasswordInputWidget`, too, replaces the value by a fixed constant
  (determined by `TAG_UNCHANGED`). When it gets this value back, it
  treats the password as unchanged. This way, you need to provide
  a password value only when you want to change it.

  Be carefull, however: the value shown on the html page has nothing to
  do with the real password value. When you want to change your password,
  you must provide a completely new value. You cannot just modify some
  characters.
  """

  TAG_UNCHANGED = "**__unchanged__**"

  __call__ = TextWidget.__call__

  def hasInput(self):
    # we trick `zope.formlib` to make the field value available by
    #  pretending to have no input when the input is our unchanged tag
    return super(PasswordInputWidget, self).hasInput() \
           and self._getFormInput() != self.TAG_UNCHANGED

  def _getFormValue(self):
    if self.hasInput(): return self._getFormInput()
    # do not pretend a value is set if it is not
    if not self._renderedValueSet() or not self._data: return ""
    # we have a non empty value -- hide it
    return self.TAG_UNCHANGED

  def _toFieldValue(self, input):
    value = super(PasswordInputWidget, self)._toFieldValue(input.strip())
    if value != self.TAG_UNCHANGED: return value
    elif self._renderedValueSet(): return self._data
    else: return self.context.missing_value


class PasswordDisplayWidget(DisplayWidget):
  __call__ = rebindFunction(DisplayWidget.__call__,
                            escape=lambda x: "<span>%s</span>" % ("*" * len(x))
                            )


##############################################################################
## Generic text widget

class GenericTextWidget(TextWidget):
  """generic TextWidget which uses the fields 'fromUnicode' to determine
  the field value.
  """
  def _toFieldValue(self, input):
    if input == self._missing:
      return self.context.missing_value
    else:
      try:
        return self.context.fromUnicode(input)
      except ValueError, v:
        raise ConversionError(_("invalid data", "invalid data"), v)



##############################################################################
## "hidden" support


def make_hidden(widget_factory, _cache={}):
  """derive a hidden widget factory from *widget_factory*.

  used in `custom_field` definitions to get a "hidden" field.
  """

  def factory(field, request):
    widget = widget_factory(field, request)
    wc = widget.__class__
    hc = _cache.get(wc)
    if hc is None:
      class Hider(wc):
        def __call__(self): return self.hidden()
      hc = _cache[wc] = Hider
    widget.__class__ = hc
    return widget
  return factory
  



# Copyright (C) 2014 by Dr. Dieter Maurer <dieter@handshake.de>
"""Auxiliaries for data schemas.

Things may not work out as you expect when used with non pure data schemas.
"""
from copy import copy
try: from itertools import izip # Python 2
except ImportError: izip = zip
from sys import version_info

from zope.interface import alsoProvides
from zope.interface.interfaces import IMethod


class DataSchemaControlled(object):
  """Base class for classes working with data schemas."""

  # to be overridden by derived classes
  _SCHEMA_ = None # the underlaying schema

  def _fields_(self):
    """the data fields associated with `__SCHEMA__`."""
    return RemoveMethods(self._SCHEMA_)


class SchemaDict(dict, DataSchemaControlled):
  """a dictionary implementing a schema.

  Be careful: methods may hide fields (not checked).
  """

  def __init__(
    self, _mapping_or_iterable_=None, \
    _schema_=None, _unknown_=None, _nodefaults_=False, \
    **kw
    ):
    """initialize the dictionary via *_mapping_or_iterable_* and *kw*.

    *_schema_* overrides `_SCHEMA_`.

    *_unknown_* should either be `None` or a mapping. If it is `None`,
    keys not defined by the associated schema result in a `KeyError`.
    Otherwise, it should be a mapping and unknown items are transfered there.

    *_nodefaults_* indicates that default values should not be provided.
    If it is false, then default values are assigned (if necessary).

    If the object has got all fields defined by the `_SCHEMA_`, then it
    is declared to provide *_SCHEMA_*. Note, we do not check that
    the object really conforms to the schema.
    """
    super(SchemaDict, self).__init__(
      *((_mapping_or_iterable_,) if _mapping_or_iterable_ is not None else ()),
      **kw
      )
    if _schema_ is not None: self.__dict__["_SCHEMA_"] = _schema_
    # check conformity
    unknown = {}; schema = self._fields_()
    for k in self.keys():
      if k not in schema:
        unknown[k] = self[k]
    for k in unknown: del self[k]
    if unknown:
      if _unknown_ is None: raise KeyError(unknown.keys())
      _unknown_.update(unknown)
    # provide default values
    if not _nodefaults_:
      for k in schema:
        if k not in self: self[k] = getattr(schema[k], "default", None)
    # check whether we implement `_SCHEMA_`
    for k in self._SCHEMA_:
      if k not in self: break
    else: alsoProvides(self, self._SCHEMA_)

  def __getattr__(self, k):
    if k in self._fields_(): return self[k]
    raise AttributeError(k)

  def __setattr__(self, k, val):
    target = self if k in self._fields_() else self.__dict__
    target[k] = val

  def copy(self): return copy(self)


class MappingBySchema(DataSchemaControlled):
  """Mixin class to provide a mapping interface for a schema.

  This is in some sense the inverse of `SchemaDict`.

  Python version compatibility note:
    The limited approach to ensure Python 2/3 compatibility
    does not support derived classes to override the mapping methods.
  """

  def iterkeys(self): return iter(self._fields_())
  def itervalues(self): return (self[k] for k in self.iterkeys())
  def iteritems(self): return ((k, self[k]) for k in self.iterkeys())
  def keys_list(self): return list(self.iterkeys())
  def values_list(self): return list(self.itervalues())
  def items_list(self): return list(self.iteritems())
  if version_info[0] >= 3: # Python 3
    keys = iterkeys; values = itervalues; items = iteritems
  else: # Python 2
    keys = keys_list; values = values_list; items = items_list
  def __contains__(self, k): return k in self._fields_()
  def update(self, mapping):
    for k, v in mapping.items(): self[k] = v
  def get(self, k, default=None):
    if k not in self: return default
    return getattr(self, k, default)
  def __getitem__(self, k):
    v = self.get(k, self)
    if v is self: raise KeyError(k)
    return v
  def __setitem__(self, k, v):
    if k not in self: raise KeyError(k)
    setattr(self, k, v)
  def __len__(self): return len(tuple(self._fields_()))




def restrict_by_schema(mapping, schema):
  """return a mapping consisting of the items of *mapping* known by *schema*."""
  return dict(i for i in mapping.iteritems() if i[0] in schema)



class RemoveMethods(object):
  """Auxiliary wrapper to filter out methods from a schema."""
  def __init__(self, schema): self.__schema = schema

  def __iter__(self):
    s = self.__schema
    return iter(k for k in s if not IMethod.providedBy(s[k]))

  def get(self, k, default=None):
    v = self.__schema.get(k, self)
    if v is self or IMethod.providedBy(v): return default
    return  v

  def __contains__(self, k):
    return self.get(k, self) is not self

  def __getitem__(self, k):
    v = self.get(k, self)
    if v is self: raise KeyError(k)
    return v

  iterkeys = __iter__
  def itervalues(self): (self[k] for k in self)
  def iteritems(self): izip(self.iterkeys(), self.itervalues)
  def keys_list(self): return list(self.iterkeys())
  def values_list(self): return list(self.itervalues())
  def items_list(self): return list(self.iteritems())
  if version_info[0] >= 3: # Python 3
    keys = iterkeys; values = itervalues; items = iteritems
  else: # Python 2
    keys = keys_list; values = values_list; items = items_list



# Copyright (C) 2011-2018 by Dr. Dieter Maurer <dieter@handshake.de>
"""Zope 2 form support for ``SchemaConfigured``."""

try: from Products.Five.formlib.formbase import EditForm, DisplayForm, PageForm
except ImportError:
  # some Zope version dropped `formlib` support from `Five`; try `five.formlib`
  from five.formlib.formbase import EditForm, DisplayForm, PageForm

from ..form import SchemaConfiguredFormMixin

# simple ZMI supporting template
from .template import form_template

class ZmiMixin(object):
  template = form_template
  support_zmi = True
  form_class = "zni-edit"


class SchemaConfiguredEditForm(SchemaConfiguredFormMixin, EditForm): pass

class SchemaConfiguredDisplayForm(SchemaConfiguredFormMixin, DisplayForm): pass

class SchemaConfiguredZmiEditForm(ZmiMixin, SchemaConfiguredEditForm): pass
class SchemaConfiguredZmiDisplayForm(ZmiMixin, SchemaConfiguredDisplayForm): pass



from os.path import abspath, dirname, join
try:
  # try to use setuptools
  from setuptools import setup
  setupArgs = dict(
      include_package_data=True,
      install_requires=['zope.schema', "dm.reuse>=2.1.1", "decorator", "six"],
      namespace_packages=['dm', 'dm.zope'],
      zip_safe=False,
      test_suite='dm.zope.schema.tests.testsuite',
      test_requires=['zope.schema'],
      )
except ImportError:
  # use distutils
  from distutils import setup
  setupArgs = dict(
    )

cd = abspath(dirname(__file__))
pd = join(cd, 'dm', 'zope', 'schema')

def pread(filename, base=pd): return open(join(base, filename)).read().rstrip()

setup(name='dm.zope.schema',
      version=pread('VERSION.txt').split('\n')[0],
      description="'zope.schema' extensions",
      long_description=pread('README.txt'),
      classifiers=[
#        'Development Status :: 3 - Alpha',
        'Development Status :: 5 - Production/Stable',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: Zope Public License',
        'Framework :: Zope',
        'Framework :: Zope :: 4',
        'Framework :: Zope2',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Topic :: Utilities',
        ],
      author='Dieter Maurer',
      author_email='dieter@handshake.de',
      url='https://pypi.org/project/dm.zope.schema',
      packages=['dm', 'dm.zope', 'dm.zope.schema'],
      keywords='application development zope schema',
      license='ZPL',
      **setupArgs
      )
